"""."""

import numpy as _np

from .device import Device as _Device
from ..search import BPMSearch as _BPMSearch


class BPMInterlock(_Device):
    """BPM Device"""

    _properties = (
        # ==============================================================
        # General
        # +++++++
        # General interlock enable:
        'IntlkEn-Sel', 'IntlkEn-Sts',
        # General interlock clear:
        'IntlkClr-Sel',
        # Minimum sum threshold enable:
        # Habilita interlock de óbita apenas quando threshold da soma
        # ultrapassar o valor em "IntlkLmtMinSum-SP"
        'IntlkMinSumEn-Sel', 'IntlkMinSumEn-Sts',
        # Minimum sum threshold (em contagens da Soma da taxa Monit1):
        'IntlkLmtMinSum-SP', 'IntlkLmtMinSum-RB',
        # Status Instantâneo:
        # Interlock instântaneo, dificilmente será detectado com
        # a implementação atual do gateware
        'Intlk-Mon',
        # Latch do interlock, limpo apenas acionando-se a PV "Clr"
        # correspondente
        'IntlkLtc-Mon',
        # ===============================================================
        # Translation (interlock de translação)
        # +++++++++++++++++++++++++++++++++++++
        # ***************************************************************
        # Condição para interlock de translação:
        #   (posição BPM downstream + posição BPM upstream)/2 > threshold
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ***************************************************************
        # Translation interlock enable:
        'IntlkTransEn-Sel', 'IntlkTransEn-Sts',
        # Translation interlock clear:
        'IntlkTransClr-Sel',
        # Thresholds (em nm da taxa Monit1):
        'IntlkLmtTransMaxX-SP', 'IntlkLmtTransMaxX-RB',
        'IntlkLmtTransMinX-SP', 'IntlkLmtTransMinX-RB',
        'IntlkLmtTransMaxY-SP', 'IntlkLmtTransMaxY-RB',
        'IntlkLmtTransMinY-SP', 'IntlkLmtTransMinY-RB',
        # Status Instantâneo:
        #   XouY mascarado pelo "Enable"
        'IntlkTransSmaller-Mon', 'IntlkTransBigger-Mon',
        'IntlkTransSmallerX-Mon', 'IntlkTransBiggerX-Mon',  # X
        'IntlkTransSmallerY-Mon', 'IntlkTransBiggerY-Mon',  # Y
        'IntlkTransSmallerAny-Mon', 'IntlkTransBiggerAny-Mon',  # X ou Y
        # limpo apenas acionando-se a PV "Clr" correspondente
        'IntlkTransSmallerLtc-Mon', 'IntlkTransBiggerLtc-Mon',
        'IntlkTransSmallerLtcX-Mon', 'IntlkTransBiggerLtcX-Mon',
        'IntlkTransSmallerLtcY-Mon', 'IntlkTransBiggerLtcY-Mon',
        # =============================================================
        # Angular (interlock de ângulo)
        # +++++++++++++++++++++++++++++
        # *************************************************************
        # Condição para interlock de ângulo:
        #   (posição BPM downstream - posição BPM upstream) > threshold
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ************************************************************
        # Anglation interlock enable:
        'IntlkAngEn-Sel', 'IntlkAngEn-Sts',
        # Anglation interlock clear:
        'IntlkAngClr-Sel',
        # Thresholds (em rad.nm da taxa Monit1.
        #  Thresholds devem ser calculados como ângulo (em rad)
        #  entre os 2 BPMs adjacentes * distância (em nm) entre eles):
        'IntlkLmtAngMaxX-SP', 'IntlkLmtAngMaxX-RB',
        'IntlkLmtAngMinX-SP', 'IntlkLmtAngMinX-RB',
        'IntlkLmtAngMaxY-SP', 'IntlkLmtAngMaxY-RB',
        'IntlkLmtAngMinY-SP', 'IntlkLmtAngMinY-RB',
        # Status Instantâneo:
        #  X ou Y mascarado pelo "Enable"
        'IntlkAngSmaller-Mon', 'IntlkAngBigger-Mon',
        'IntlkAngSmallerAny-Mon', 'IntlkAngBiggerAny-Mon',  # X ou Y
        'IntlkAngSmallerX-Mon', 'IntlkAngBiggerX-Mon',  # X
        'IntlkAngSmallerY-Mon', 'IntlkAngBiggerY-Mon',  # Y
        # limpo apenas acionando-se a PV "Clr" correspondente
        'IntlkAngSmallerLtc-Mon', 'IntlkAngBiggerLtc-Mon',
        'IntlkAngSmallerLtcX-Mon', 'IntlkAngBiggerLtcX-Mon',
        'IntlkAngSmallerLtcY-Mon', 'IntlkAngBiggerLtcY-Mon',
        # ============================================================
        )

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    def __init__(self, devname):
        """."""
        # call base class constructor
        if not _BPMSearch.is_valid_devname(devname):
            raise ValueError(devname + ' is no a valid BPM or PBPM name.')
        super().__init__(devname, properties=BPMInterlock._properties)

    @property
    def enabled(self):
        """.General interlock enable."""
        return self['IntlkEn-Sts']

    @enabled.setter
    def enabled(self, value):
        self['IntlkEn-Sel'] = int(value)

    def cmd_reset(self):
        """General interlock clear."""
        self['IntlkClr-Sel'] = 1
        return True



"""."""
import time as _time

import numpy as _np

from .device import Device as _Device, ProptyDevice as _ProptyDevice
from ..timesys.csdev import ETypes as _ETypes, Const as _TIConst, \
    get_hl_trigger_database as _get_hl_trigger_database
from ..util import get_bit as _get_bit


class EVG(_Device):
    """Device EVG."""

    DEVNAME = 'AS-RaMO:TI-EVG'

    _properties = (
        'InjectionEvt-Sel', 'InjectionEvt-Sts', 'UpdateEvt-Cmd',
        'ContinuousEvt-Sel', 'ContinuousEvt-Sts',
        'RepeatBucketList-SP', 'RepeatBucketList-RB',
        'BucketList-SP', 'BucketList-RB', 'BucketList-Mon',
        'BucketListLen-Mon', 'TotalInjCount-Mon', 'InjCount-Mon',
        'BucketListSyncStatus-Mon')

    def __init__(self):
        """."""
        super().__init__(EVG.DEVNAME, properties=EVG._properties)

    @property
    def nrpulses(self):
        """Number of pulses to repeat Bucket List."""
        return self['RepeatBucketList-RB']

    @nrpulses.setter
    def nrpulses(self, value):
        """Set number of pulses to repeat Bucket List."""
        self['RepeatBucketList-SP'] = bool(value)

    @property
    def bucketlist_len(self):
        """Bucket List Length."""
        value = self['BucketListLen-Mon']
        return int(value) if value is not None else None

    @property
    def bucketlist_mon(self):
        """Implemented Bucket List."""
        return self['BucketList-Mon'][:self.bucketlist_len]

    @property
    def bucketlist(self):
        """Last setpoint accepted for Bucket List."""
        return self['BucketList-RB'][:self.bucketlist_len]

    @bucketlist.setter
    def bucketlist(self, value):
        """Set Bucket List."""
        self['BucketList-SP'] = _np.array(value, dtype=int)

    @property
    def bucketlist_sync(self):
        """Bucket list syncronized status."""
        return self['BucketListSyncStatus-Mon']

    @property
    def continuous_state(self):
        """."""
        return self['ContinuousEvt-Sts']

    @continuous_state.setter
    def continuous_state(self, value):
        """."""
        self['ContinuousEvt-Sel'] = bool(value)

    @property
    def injection_state(self):
        """."""
        return self['InjectionEvt-Sts']

    @injection_state.setter
    def injection_state(self, value):
        """."""
        self['InjectionEvt-Sel'] = bool(value)

    @property
    def injection_count_total(self):
        """."""
        return self['TotalInjCount-Mon']

    @property
    def injection_count(self):
        """."""
        return self['InjCount-Mon']

    def fill_bucketlist(self, stop, start=1, step=30, timeout=10):
        """Fill bucket list."""
        if (step > 0) and (start > stop):
            stop += 864
        if (step < 0) and (stop > start):
            stop -= 864
        value = _np.arange(start, stop, step)
        value = (value-1) % 864 + 1
        self.bucketlist = value
        rb_value = _np.zeros(864)
        rb_value[:len(value)] = value
        return self._wait('BucketList-RB', rb_value, timeout=timeout)

    def wait_injection_finish(self, timeout=10):
        """."""
        return self._wait(propty='InjectionEvt-Sts', value=0, timeout=timeout)

    def cmd_update_events(self, timeout=10):
        """."""
        val = self.continuous_state
        self['UpdateEvt-Cmd'] = 1
        _time.sleep(0.1)
        return self._wait(
            propty='ContinuousEvt-Sts', value=val, timeout=timeout)

    def cmd_turn_on_injection(self, timeout=10):
        """."""
        self.injection_state = 1
        return self._wait(propty='InjectionEvt-Sel', value=1, timeout=timeout)

    def cmd_turn_off_injection(self, timeout=10):
        """."""
        self.injection_state = 0
        return self._wait(propty='InjectionEvt-Sel', value=0, timeout=timeout)

    def cmd_turn_on_continuous(self, timeout=10):
        """."""
        self.continuous_state = 1
        return self._wait(propty='ContinuousEvt-Sel', value=1, timeout=timeout)

    def cmd_turn_off_continuous(self, timeout=10):
        """."""
        self.continuous_state = 0
        return self._wait(propty='ContinuousEvt-Sel', value=0, timeout=timeout)

    def set_nrpulses(self, value, timeout=10):
        """Set and wait number of pulses."""
        self['RepeatBucketList-SP'] = value
        return self._wait('RepeatBucketList-RB', value, timeout=timeout)


class Event(_ProptyDevice):
    """Device Timing Event."""

    _properties = (
        'Delay-SP', 'Delay-RB', 'DelayRaw-SP', 'DelayRaw-RB',
        'DelayType-Sel', 'DelayType-Sts', 'Mode-Sel', 'Mode-Sts',
        'Code-Mon', 'ExtTrig-Cmd',
        )

    MODES = _ETypes.EVT_MODES
    DELAYTYPES = ('Incr', 'Fixed')

    def __init__(self, evtname):
        """."""
        super().__init__(
            EVG.DEVNAME, evtname, properties=Event._properties)

    @property
    def mode(self):
        """."""
        return self['Mode-Sts']

    @mode.setter
    def mode(self, value):
        self._enum_setter('Mode-Sel', value, Event.MODES)

    @property
    def mode_str(self):
        """."""
        return Event.MODES[self['Mode-Sts']]

    @property
    def code(self):
        """."""
        return self['Code-Mon']

    @property
    def delay_type(self):
        """."""
        return self['DelayType-Sts']

    @delay_type.setter
    def delay_type(self, value):
        self._enum_setter('DelayType-Sel', value, Event.DELAYTYPES)

    @property
    def delay_type_str(self):
        """."""
        return Event.DELAYTYPES[self['DelayType-Sts']]

    @property
    def delay(self):
        """."""
        return self['Delay-RB']

    @delay.setter
    def delay(self, value):
        self['Delay-SP'] = value

    @property
    def delay_raw(self):
        """."""
        return self['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        self['DelayRaw-SP'] = int(value)

    def cmd_external_trigger(self):
        """."""
        self['ExtTrig-Cmd'] = 1
        return True

    @property
    def is_in_inj_table(self):
        """."""
        return self.mode_str in Event.MODES[1:4]


class Trigger(_Device):
    """Device trigger."""

    STATES = ('Dsbl', 'Enbl')
    LOCKLL = ('Unlocked', 'Locked')
    POLARITIES = ('Normal', 'Inverse')

    def __init__(self, trigname):
        """Init."""
        self._database = _get_hl_trigger_database(trigname)
        self._properties = tuple(self._database)
        self._source_options = self._database['Src-Sel']['enums']
        super().__init__(trigname, properties=self._properties)

    @property
    def status(self):
        """Status."""
        return self['Status-Mon']

    @property
    def status_str(self):
        """Status string."""
        value = self.status
        strs = [d for i, d in enumerate(_TIConst.HLTrigStatusLabels)
                if _get_bit(value, i)]
        return ', '.join(strs) if strs else 'Ok'

    @property
    def status_labels(self):
        """Return Status labels of trigger.

        Returns:
            list: labels describing the possible statuses of the trigger.

        """
        pvo = self.pv_object('StatusLabels-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def state(self):
        """State."""
        return self['State-Sts']

    @state.setter
    def state(self, value):
        self._enum_setter('State-Sel', value, Trigger.STATES)

    @property
    def state_str(self):
        """State string."""
        return Trigger.STATES[self['State-Sts']]

    @property
    def lock_low_level(self):
        """Lock low level status."""
        return self['LowLvlLock-Sts']

    @lock_low_level.setter
    def lock_low_level(self, value):
        self._enum_setter('LowLvlLock-Sel', value, Trigger.LOCKLL)

    @property
    def lock_low_level_str(self):
        """Lock low level status string."""
        return Trigger.LOCKLL[self['LowLvlLock-Sts']]

    @property
    def controlled_channels(self):
        """Return channels controlled by this trigger.

        Returns:
            list: names of the controlled channels.

        """
        pvo = self.pv_object('CtrldChannels-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def low_level_triggers(self):
        """Return low level triggers controlled by this trigger.

        Returns:
            list: names of the controlled low level triggers.

        """
        pvo = self.pv_object('LowLvlTriggers-Cte')
        return pvo.get(as_string=True).split('\n')

    @property
    def source(self):
        """Source."""
        return self['Src-Sts']

    @source.setter
    def source(self, value):
        self._enum_setter('Src-Sel', value, self._source_options)

    @property
    def source_str(self):
        """Source string."""
        return self._source_options[self['Src-Sts']]

    @property
    def source_options(self):
        """Source options."""
        return self._source_options

    @property
    def duration(self):
        """Duration."""
        return self['Duration-RB']

    @duration.setter
    def duration(self, value):
        self['Duration-SP'] = value

    @property
    def polarity(self):
        """Polarity."""
        return self['Polarity-Sts']

    @polarity.setter
    def polarity(self, value):
        self._enum_setter('Polarity-Sel', value, Trigger.POLARITIES)

    @property
    def polarity_str(self):
        """Polarity string."""
        return Trigger.POLARITIES[self['Polarity-Sts']]

    @property
    def nr_pulses(self):
        """Nr. of pulses."""
        return self['NrPulses-RB']

    @nr_pulses.setter
    def nr_pulses(self, value):
        self['NrPulses-SP'] = value

    @property
    def delay(self):
        """Delay."""
        return self['Delay-RB']

    @delay.setter
    def delay(self, value):
        self['Delay-SP'] = value

    @property
    def total_delay(self):
        """Total delay."""
        return self['TotalDelay-Mon']

    @property
    def delta_delay(self):
        """Return delta delay array.

        Returns:
            numpy.ndarray: delta delays.

        """
        return self['DeltaDelay-RB']

    @delta_delay.setter
    def delta_delay(self, value):
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise TypeError('Value must be a numpy.ndarray, list or tuple.')

        value = _np.array(value)
        size = self.delta_delay.size
        if value.size != size:
            raise TypeError(f'Size of value must be {size:d}.')
        self['DeltaDelay-SP'] = value

    @property
    def delay_raw(self):
        """Delay raw."""
        return self['DelayRaw-RB']

    @delay_raw.setter
    def delay_raw(self, value):
        self['DelayRaw-SP'] = value

    @property
    def total_delay_raw(self):
        """Total delay raw."""
        return self['TotalDelayRaw-Mon']

    @property
    def delta_delay_raw(self):
        """Return delta delay raw array.

        Returns:
            numpy.ndarray: delta delays raw.

        """
        return self['DeltaDelayRaw-RB']

    @delta_delay_raw.setter
    def delta_delay_raw(self, value):
        if not isinstance(value, (_np.ndarray, list, tuple)):
            raise TypeError('Value must be a numpy.ndarray, list or tuple.')

        value = _np.array(value)
        size = self.delta_delay_raw.size
        if value.size != size:
            raise TypeError(f'Size of value must be {size:d}.')
        self['DeltaDelayRaw-SP'] = value

    @property
    def is_in_inj_table(self):
        """Is in Injection table."""
        return self['InInjTable-Mon']

    def cmd_enable(self, timeout=3):
        """Command enable."""
        self.state = 1
        return self._wait('State-Sts', 1, timeout)

    def cmd_disable(self, timeout=3):
        """Command disable."""
        self.state = 0
        return self._wait('State-Sts', 0, timeout)

    def cmd_lock_low_level(self, timeout=3):
        """Lock low level IOCs state."""
        self.lock_low_level = 1
        return self._wait('LowLvlLock-Sts', 1, timeout)

    def cmd_unlock_low_level(self, timeout=3):
        """Unlock low level IOCs state."""
        self.lock_low_level = 0
        return self._wait('LowLvlLock-Sts', 0, timeout)



"""."""

from .device import Device as _Device


class Screen(_Device):
    """."""

    class DEVICES:
        """Devices names."""

        TB_1 = 'TB-01:DI-Scrn-1'
        TB_2 = 'TB-01:DI-Scrn-2'
        TB_3 = 'TB-02:DI-Scrn-1'
        TB_4 = 'TB-02:DI-Scrn-2'
        TB_5 = 'TB-03:DI-Scrn'
        TB_6 = 'TB-04:DI-Scrn'
        BO_1 = 'BO-01D:DI-Scrn-1'
        BO_2 = 'BO-01D:DI-Scrn-2'
        BO_3 = 'BO-02U:DI-Scrn'
        TS_1 = 'TS-01:DI-Scrn'
        TS_2 = 'TS-02:DI-Scrn'
        TS_3 = 'TS-03:DI-Scrn'
        TS_4 = 'TS-04:DI-Scrn-1'
        TS_5 = 'TS-04:DI-Scrn-2'
        TS_6 = 'TS-04:DI-Scrn-3'
        TB = (TB_1, TB_2, TB_3, TB_4, TB_5, TB_6)
        TS = (TS_1, TS_2, TS_3, TS_4, TS_5, TS_6)
        BO = (BO_1, BO_2, BO_3)
        ALL = (
            TB_1, TB_2, TB_3, TB_4, TB_5, TB_6,
            BO_1, BO_2, BO_3,
            TS_1, TS_2, TS_3, TS_4, TS_5, TS_6)

    _properties = (
        'ImgData-Mon',
        'CenterXDimFei-Mon', 'CenterYDimFei-Mon',
        'SigmaXDimFei-Mon', 'SigmaYDimFei-Mon',
        'ThetaDimFei-Mon',
    )

    def __init__(self, devname):
        """."""
        # check if device exists
        if devname not in Screen.DEVICES.ALL:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=Screen._properties)

    @property
    def image(self):
        """."""
        return self['ImgData-Mon']

    @property
    def centerx(self):
        """."""
        return self['CenterXDimFei-Mon']

    @property
    def centery(self):
        """."""
        return self['CenterYDimFei-Mon']

    @property
    def sigmax(self):
        """."""
        return self['SigmaXDimFei-Mon']

    @property
    def sigmay(self):
        """."""
        return self['SigmaYDimFei-Mon']

    @property
    def angle(self):
        """."""
        return self['ThetaDimFei-Mon']



"""Injection Control App."""

from .. import csdev as _csdev
from ..devices import InjSysStandbyHandler as _InjSysHandler, \
    EGun as _EGun


# --- Enumeration Types ---

class ETypes(_csdev.ETypes):
    """Local enumerate types."""

    INJMODE = ('Decay', 'TopUp')
    INJTYPE_SEL = ('SingleBunch', 'MultiBunch')
    INJTYPE_STS = ('SingleBunch', 'MultiBunch', 'Undefined')
    TOPUPSTS = (
        'Off', 'Waiting', 'TurningOn', 'Injecting', 'TurningOff')
    INJSYSCMDSTS = ('Idle', 'On', 'Off')


_et = ETypes


# --- Const class ---

class Const(_csdev.Const):
    """Const class."""

    InjMode = _csdev.Const.register('InjMode', _et.INJMODE)
    InjTypeSel = _csdev.Const.register('InjTypeSel', _et.INJTYPE_SEL)
    InjTypeSts = _csdev.Const.register('InjTypeSts', _et.INJTYPE_STS)
    TopUpSts = _csdev.Const.register('TopUpSts', _et.TOPUPSTS)
    InjSysCmdSts = _csdev.Const.register('InjSysCmdSts', _et.INJSYSCMDSTS)

    GEN_STATUS_LABELS = ('LI', 'TB', 'BO', 'TS', 'SI', 'AS')
    LI_STATUS_LABELS = ('Egun', 'PS', 'PU', 'RF', 'TI')
    TL_STATUS_LABELS = ('PS', 'PU', 'TI')
    SR_STATUS_LABELS = ('PS', 'PU', 'RF', 'TI')
    AS_STATUS_LABELS = ('TI', )
    INJ_STATUS_LABELS = (
        'TI ContinuousEvt is off',
        'BucketList not synced',
        'EGBiasPS voltage diff. from desired',
        'EGFilaPS current diff. from op.value',
        'EGHVPS voltage diff. from op.value',
        'EGPulsePS setup is diff. from desired',
        'EGTriggerPS is off',
        'Inj.System is not on',
    )

    MIN_BKT = 1
    MAX_BKT = 864

    RF_RMP_TIMEOUT = 3*60  # [s]
    TI_INJ_TIMEOUT = 3*60  # [s]


_ct = Const


def get_status_labels(sec=''):
    """Return Status Labels enum."""
    if sec == 'LI':
        return _ct.LI_STATUS_LABELS
    if sec in ['TB', 'TS']:
        return _ct.TL_STATUS_LABELS
    if sec in ['BO', 'SI']:
        return _ct.SR_STATUS_LABELS
    if sec == 'AS':
        return _ct.AS_STATUS_LABELS
    if sec == 'Inj':
        return _ct.INJ_STATUS_LABELS
    return _ct.GEN_STATUS_LABELS


# --- Databases ---


def get_injctrl_propty_database():
    """Return property database of injection control IOC."""
    # injsys properties
    injsys_onorder = ','.join(_InjSysHandler.DEF_ON_ORDER)
    injsys_offorder = ','.join(_InjSysHandler.DEF_OFF_ORDER)
    # egun properties
    egsbbias = _EGun.BIAS_SINGLE_BUNCH
    egmbbias = _EGun.BIAS_MULTI_BUNCH
    egfilacurr = _EGun.FILACURR_OPVALUE
    eghvolt = _EGun.HV_OPVALUE

    dbase = {
        'Version-Cte': {'type': 'str', 'value': 'UNDEF'},
        'Log-Mon': {'type': 'str', 'value': 'Starting...'},

        'Mode-Sel': {
            'type': 'enum', 'value': _ct.InjMode.Decay,
            'enums': _et.INJMODE},
        'Mode-Sts': {
            'type': 'enum', 'value': _ct.InjMode.Decay,
            'enums': _et.INJMODE},
        'Type-Sel': {
            'type': 'enum', 'value': _ct.InjTypeSel.MultiBunch,
            'enums': _et.INJTYPE_SEL},
        'Type-Sts': {
            'type': 'enum', 'value': _ct.InjTypeSts.Undefined,
            'enums': _et.INJTYPE_STS},
        'SglBunBiasVolt-SP': {
            'type': 'float', 'value': egsbbias, 'prec': 1,
            'unit': 'V', 'lolim': -150.0, 'hilim': 0.0},
        'SglBunBiasVolt-RB': {
            'type': 'float', 'value': egsbbias, 'prec': 1,
            'unit': 'V', 'lolim': -150.0, 'hilim': 0.0},
        'MultBunBiasVolt-SP': {
            'type': 'float', 'value': egmbbias, 'prec': 1,
            'unit': 'V', 'lolim': -150.0, 'hilim': 0.0},
        'MultBunBiasVolt-RB': {
            'type': 'float', 'value': egmbbias, 'prec': 1,
            'unit': 'V', 'lolim': -150.0, 'hilim': 0.0},
        'FilaOpCurr-SP': {
            'type': 'float', 'value': egfilacurr, 'prec': 3,
            'unit': 'A', 'lolim': 0.0, 'hilim': 1.5},
        'FilaOpCurr-RB': {
            'type': 'float', 'value': egfilacurr, 'prec': 3,
            'unit': 'A', 'lolim': 0.0, 'hilim': 1.5},
        'HVOpVolt-SP': {
            'type': 'float', 'value': eghvolt, 'prec': 3,
            'unit': 'kV', 'lolim': 0.0, 'hilim': 95.0},
        'HVOpVolt-RB': {
            'type': 'float', 'value': eghvolt, 'prec': 3,
            'unit': 'kV', 'lolim': 0.0, 'hilim': 95.0},
        'TargetCurrent-SP': {
            'type': 'float', 'value': 100.0, 'unit': 'mA',
            'prec': 2, 'lolim': 0.0, 'low': 0.0, 'lolo': 0.0,
            'hilim': 500.0, 'high': 500.0, 'hihi': 500.0},
        'TargetCurrent-RB': {
            'type': 'float', 'value': 100.0, 'unit': 'mA',
            'prec': 2, 'lolim': 0.0, 'low': 0.0, 'lolo': 0.0,
            'hilim': 500.0, 'high': 500.0, 'hihi': 500.0},
        'BucketListStart-SP': {
            'type': 'int', 'value': 1, 'unit': 'bucket index',
            'lolim': _ct.MIN_BKT, 'hilim': _ct.MAX_BKT},
        'BucketListStart-RB': {
            'type': 'int', 'value': 1, 'unit': 'bucket index',
            'lolim': _ct.MIN_BKT, 'hilim': _ct.MAX_BKT},
        'BucketListStop-SP': {
            'type': 'int', 'value': 864, 'unit': 'bucket index',
            'lolim': _ct.MIN_BKT, 'hilim': _ct.MAX_BKT, },
        'BucketListStop-RB': {
            'type': 'int', 'value': 864, 'unit': 'bucket index',
            'lolim': _ct.MIN_BKT, 'hilim': _ct.MAX_BKT},
        'BucketListStep-SP': {
            'type': 'int', 'value': 15, 'unit': 'buckets',
            'lolim': -_ct.MAX_BKT+1, 'hilim': _ct.MAX_BKT-1},
        'BucketListStep-RB': {
            'type': 'int', 'value': 15, 'unit': 'buckets',
            'lolim': -_ct.MAX_BKT+1, 'hilim': _ct.MAX_BKT-1},

        'TopUpState-Sel': {
            'type': 'enum', 'value': _ct.OffOn.Off, 'enums': _et.OFF_ON},
        'TopUpState-Sts': {
            'type': 'enum', 'value': _ct.TopUpSts.Off, 'enums': _et.TOPUPSTS},
        'TopUpPeriod-SP': {
            'type': 'int', 'value': 15*60, 'unit': 's',
            'lolim': 30, 'hilim': 6*60*60},
        'TopUpPeriod-RB': {
            'type': 'int', 'value': 15*60, 'unit': 's',
            'lolim': 30, 'hilim': 6*60*60},
        'TopUpNextInj-Mon': {
            'type': 'float', 'value': 0.0, 'unit': 's'},
        'TopUpNextInjRound-Cmd': {'type': 'int', 'value': 0},
        'TopUpMaxNrPulses-SP': {
            'type': 'int', 'value': 100, 'unit': 'pulses',
            'lolim': 1, 'hilim': 1000},
        'TopUpMaxNrPulses-RB': {
            'type': 'int', 'value': 100, 'unit': 'pulses',
            'lolim': 1, 'hilim': 1000},
        'AutoStop-Sel': {
            'type': 'enum', 'value': _ct.OffOn.Off, 'enums': _et.OFF_ON},
        'AutoStop-Sts': {
            'type': 'enum', 'value': _ct.OffOn.Off, 'enums': _et.OFF_ON},

        'InjSysTurnOn-Cmd': {'type': 'int', 'value': 0},
        'InjSysTurnOff-Cmd': {'type': 'int', 'value': 0},
        'InjSysCmdDone-Mon': {
            'type': 'string', 'value': ''},
        'InjSysCmdSts-Mon': {
            'type': 'enum', 'value': _ct.InjSysCmdSts.Idle,
            'enums': _et.INJSYSCMDSTS},
        'InjSysTurnOnOrder-SP': {'type': 'string', 'value': injsys_onorder},
        'InjSysTurnOnOrder-RB': {'type': 'string', 'value': injsys_onorder},
        'InjSysTurnOffOrder-SP': {'type': 'string', 'value': injsys_offorder},
        'InjSysTurnOffOrder-RB': {'type': 'string', 'value': injsys_offorder},

        'DiagStatusLI-Mon': {
            'type': 'int', 'value': 2**len(_ct.LI_STATUS_LABELS)-1},
        'DiagStatusLILabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.LI_STATUS_LABELS)},
        'DiagStatusTB-Mon': {
            'type': 'int', 'value': 2**len(_ct.TL_STATUS_LABELS)-1},
        'DiagStatusTBLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.TL_STATUS_LABELS)},
        'DiagStatusBO-Mon': {
            'type': 'int', 'value': 2**len(_ct.SR_STATUS_LABELS)-1},
        'DiagStatusBOLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.SR_STATUS_LABELS)},
        'DiagStatusTS-Mon': {
            'type': 'int', 'value': 2**len(_ct.TL_STATUS_LABELS)-1},
        'DiagStatusTSLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.TL_STATUS_LABELS)},
        'DiagStatusSI-Mon': {
            'type': 'int', 'value': 2**len(_ct.SR_STATUS_LABELS)-1},
        'DiagStatusSILabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.SR_STATUS_LABELS)},
        'DiagStatusAS-Mon': {
            'type': 'int', 'value': 2**len(_ct.AS_STATUS_LABELS)-1},
        'DiagStatusASLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.AS_STATUS_LABELS)},
        'DiagStatus-Mon': {
            'type': 'int', 'value': 2**len(_ct.GEN_STATUS_LABELS)-1},
        'DiagStatusLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.GEN_STATUS_LABELS)},
        'InjStatus-Mon': {
            'type': 'int', 'value': 2**len(_ct.INJ_STATUS_LABELS)-1},
        'InjStatusLabels-Cte': {
            'type': 'char', 'count': 1000,
            'value': '\n'.join(_ct.INJ_STATUS_LABELS)},
    }
    dbase = _csdev.add_pvslist_cte(dbase)
    return dbase
